# %% imports
import autogaita
import autogaita.gui.gaita_widgets as gaita_widgets
import autogaita.gui.gui_utils as gui_utils
import tkinter as tk
import customtkinter as ctk
import pandas as pd
import os
import math
from threading import Thread
import platform
import json
import copy


# %% global constants
from autogaita.gui.gui_constants import (
    GROUP_FG_COLOR,
    GROUP_HOVER_COLOR,
    HEADER_FONT_NAME,
    HEADER_FONT_SIZE,
    HEADER_TXT_COLOR,
    MAIN_HEADER_FONT_SIZE,
    TEXT_FONT_NAME,
    TEXT_FONT_SIZE,
    ADV_CFG_TEXT_FONT_SIZE,
    COLOR_PALETTES_LIST,
    WINDOWS_TASKBAR_MAXHEIGHT,
    AUTOGAITA_FOLDER_PATH,
    get_widget_cfg_dict,  # function!
)
from autogaita.group.group_constants import NORM_SHEET_NAME, AVG_GROUP_SHEET_NAME

# these colors are GUI-specific - add to common widget cfg
FG_COLOR = GROUP_FG_COLOR
HOVER_COLOR = GROUP_HOVER_COLOR
WIDGET_CFG = get_widget_cfg_dict()
WIDGET_CFG["FG_COLOR"] = FG_COLOR
WIDGET_CFG["HOVER_COLOR"] = HOVER_COLOR

# group GUI specific constants
MIN_GROUP_NUM = 2
MAX_GROUP_NUM = 6
CONFIG_FILE_NAME = "group_gui_config.json"
STRING_VARS = ["group_names", "group_dirs", "results_dir", "load_dir"]
# note n_components can be float only for this input-check since we convert to int when
# fitting the PCA model if it is equal to or greater than 1
FLOAT_VARS = ["stats_threshold", "PCA_n_components"]
LIST_VARS = [
    "stats_variables",  #  stats/PCA variables are also TK_BOOL_VARS but this will be
    "PCA_variables",  #  handled within the ---PCA / STATS FEATURE FRAMES--- part
]
INT_VARS = ["permutation_number"]
# TK_BOOL/STR_VARS are only used for initialising widgets based on cfg file
# (note that numbers are initialised as strings)
TK_BOOL_VARS = [
    "do_permtest",
    "do_anova",
    "PCA_save_3D_video",
    "plot_SE",
    "legend_outside",
    "dont_show_plots",
]
TK_STR_VARS = [
    "anova_design",
    "permutation_number",
    "stats_threshold",
    "PCA_n_components",
    "PCA_custom_scatter_PCs",
    "PCA_bins",
    "which_leg",
    "results_dir",
    "load_dir",
    "color_palette",
]
EXCLUDED_VARS_FROM_CFG_FILE = ["last_runs_stats_variables", "last_runs_PCA_variables"]

# %%...............................  MAIN PROGRAM ......................................


def run_group_gui():
    # ..................................................................................
    # .....................  root (intro) window initialisation  .......................
    # ..................................................................................
    # Check for config file
    config_file_path = os.path.join(AUTOGAITA_FOLDER_PATH, CONFIG_FILE_NAME)
    if not os.path.isfile(config_file_path):
        config_file_error_msg = (
            "group_gui_config.json file not found in autogaita folder.\n"
            "Confirm that the file exists and is named correctly.\n"
            "If not, download it again from the GitHub repository."
        )
        tk.messagebox.showerror(
            title="Config File Error", message=config_file_error_msg
        )
        exit()

    # CustomTkinter vars
    ctk.set_appearance_mode("dark")  # Modes: system (default), light, dark
    ctk.set_default_color_theme("green")  # Themes: blue , dark-blue, green
    # root
    root = ctk.CTk()
    fix_window_after_its_creation(root)
    gui_utils.configure_the_icon(root)

    # prepare screen width / height for later windows
    global screen_width, screen_height
    screen_width = root.winfo_screenwidth()  # width of the screen
    screen_height = root.winfo_screenheight()  # height of the screen
    if platform.system() == "Windows":  # adjust for taskbar in windows only
        screen_height -= WINDOWS_TASKBAR_MAXHEIGHT
    root_w = screen_width / 2  # width for the Tk root
    root_h = screen_height / 4.5  # height for the Tk root
    # calculate x and y coordinates for the Tk root window
    root_x = (screen_width / 2) - (root_w / 2)
    root_y = (screen_height / 2) - (root_h / 1)
    root_dimensions = (root_w, root_h, root_x, root_y)
    # set the dimensions of the screen and where it is placed
    root.geometry("%dx%d+%d+%d" % root_dimensions)
    root.title("Group GaitA")

    # ..................................................................................
    # .......................  root (intro) window population  .........................
    # ..................................................................................

    # .................................  widgets  ......................................
    # welcome message
    welcomestring = "Welcome to AutoGaitA Group!"
    welcomeheader_label = ctk.CTkLabel(
        root,
        text=welcomestring,
        width=root_w,
        fg_color=FG_COLOR,
        text_color=HEADER_TXT_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
    )
    welcomeheader_label.grid(row=0, column=0, columnspan=MAX_GROUP_NUM - 1)
    # contrast number message
    numberstring = "How many groups would you like to compare with each other?"
    numberheader_label = ctk.CTkLabel(
        root,
        text=numberstring,
        width=root_w,
        text_color=HEADER_TXT_COLOR,
        font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
    )
    numberheader_label.grid(row=1, column=0, columnspan=MAX_GROUP_NUM - 1, pady=(20, 5))
    # number of groups to contrast radio buttons
    group_number = ctk.IntVar(value=0)
    group_buttons = []
    for i in range(MIN_GROUP_NUM, MAX_GROUP_NUM + 1):
        group_buttons.append(
            ctk.CTkRadioButton(
                root,
                text=str(i),
                variable=group_number,
                value=i,
                fg_color=FG_COLOR,
                hover_color=HOVER_COLOR,
                font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
            )
        )
        group_buttons[-1].grid(row=2, column=i - MIN_GROUP_NUM)
    # continue button
    continue_button = ctk.CTkButton(
        root,
        text="Continue",
        fg_color=FG_COLOR,
        hover_color=HOVER_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
        command=lambda: (
            build_mainwindow(root, group_number, root_dimensions),
            root.withdraw(),
        ),
    )
    continue_button.grid(
        row=3,
        column=0,
        columnspan=MAX_GROUP_NUM - 1,
        pady=23,
        padx=30,
        sticky="nsew",
    )
    root.grid_rowconfigure(3, weight=1)
    num_cols = root.grid_size()[0]  # maximise cols
    for c in range(num_cols):
        root.grid_columnconfigure(c, weight=1)

    root.mainloop()


# %%..................  LOCAL FUNCTION(S) #1 - BUILD MAIN WINDOW  ......................
def build_mainwindow(root, group_number, root_dimensions):
    """Build the main window based on the number of groups to compare"""
    group_number = group_number.get()
    if group_number == 0:
        error_message = "You did not specify a number of groups for your contrast."
        tk.messagebox.showerror(title="No Input", message=error_message)
        root.deiconify()
    else:
        # .....................  important - global cfg variable  ......................
        # we have cfg be global so it can be modified by all widgets and frames.
        # just prior to calling autogaita.group, we will copy its values to a temporary # cfg var, so the values of the global variable are never used for running
        # anything (see run_analysis)
        global cfg
        cfg = extract_cfg_from_json_file(root)

        # ........................  geometry & intro section  ..........................
        # geometry
        mainwindow = ctk.CTkToplevel(root)
        mainwindow.title("Group GaitA")
        # set the dimensions of the screen and where it is placed
        # => have it half-wide starting at 1/4 of screen's width (dont change w & x!)
        mainwindow.geometry(
            f"{int(screen_width / 2)}x{screen_height}+{int(screen_width / 4)}+0"
        )
        fix_window_after_its_creation(mainwindow)

        # ..........................  main configuration  ..............................
        # config header
        cfgheader_label = ctk.CTkLabel(
            mainwindow,
            text="Main Configuration",
            width=screen_width,
            fg_color=FG_COLOR,
            text_color=HEADER_TXT_COLOR,
            font=(HEADER_FONT_NAME, MAIN_HEADER_FONT_SIZE),
        )
        cfgheader_label.grid(row=0, column=0, columnspan=3, sticky="nsew")
        # load group names and dirs from the config file
        initial_names, initial_dirs = extract_group_names_and_dirs_from_json_file()
        # group names & dirs - variables
        group_names = [
            tk.StringVar(mainwindow, initial_names[g]) for g in range(group_number)
        ]
        group_dirs = [
            tk.StringVar(mainwindow, initial_dirs[g]) for g in range(group_number)
        ]
        group_names_labels = []
        group_names_entries = []
        group_dirs_entries = []
        # group number, name & dir labels
        group_number_label = ctk.CTkLabel(
            mainwindow,
            text="Group Number",
            font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
        )
        group_number_label.grid(row=1, column=0)
        group_name_label = ctk.CTkLabel(
            mainwindow,
            text="Name",
            font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
        )
        group_name_label.grid(row=1, column=1)
        group_dir_label = ctk.CTkLabel(
            mainwindow,
            text="Path to first-level Results",
            font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
        )
        group_dir_label.grid(row=1, column=2)
        row_counter = 2  # track row idxs for loop-widget-creation
        # ...........................  group-widgets-loop  .............................
        for g in range(group_number):
            this_groups_number = g + 1
            # group numbers - labels
            group_names_labels.append(
                ctk.CTkLabel(
                    mainwindow,
                    text=str(this_groups_number),
                    font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
                )
            )
            group_names_labels[-1].grid(row=row_counter, column=0)
            # group names - entries
            group_names_entries.append(
                ctk.CTkEntry(
                    mainwindow,
                    textvariable=group_names[g],
                    font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
                )
            )
            group_names_entries[-1].grid(row=row_counter, column=1)
            # group dirs - entries
            group_dirs_entries.append(
                ctk.CTkEntry(
                    mainwindow,
                    textvariable=group_dirs[g],
                    font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
                )
            )
            group_dirs_entries[-1].grid(row=row_counter, column=2, sticky="ew")
            # update row counter
            row_counter += 1
        # call row_counter a better name
        last_group_row = row_counter
        # empty label 1 for spacing
        empty_label_one = ctk.CTkLabel(mainwindow, text="")
        empty_label_one.grid(row=last_group_row, column=0, columnspan=3, sticky="nsew")

        # NOTE! load results and load dirs as strings from cfg json before converting
        #       to tk-vars
        results_dir_string, load_dir_string = (
            extract_results_and_load_dirs_from_json_files()
        )

        # results dir
        results_dir = tk.StringVar(mainwindow, results_dir_string)
        results_dir_label, results_dir_entry = gaita_widgets.label_and_entry_pair(
            mainwindow,
            "Path to save group-analysis results to",
            results_dir,
            WIDGET_CFG,
        )
        results_dir_label.grid(
            row=last_group_row + 1, column=0, columnspan=3, sticky="ew"
        )
        results_dir_entry.grid(
            row=last_group_row + 2, column=0, columnspan=3, sticky="ew"
        )
        # load dir
        load_dir = tk.StringVar(mainwindow, load_dir_string)
        load_dir_label, load_dir_entry = gaita_widgets.label_and_entry_pair(
            mainwindow,
            "Optional: path to load previous group-data from",
            load_dir,
            WIDGET_CFG,
        )
        load_dir_label.grid(row=last_group_row + 3, column=0, columnspan=3, sticky="ew")
        load_dir_entry.grid(row=last_group_row + 4, column=0, columnspan=3, sticky="ew")
        # empty label 2 for spacing
        empty_label_two = ctk.CTkLabel(mainwindow, text="")
        empty_label_two.grid(
            row=last_group_row + 5, column=0, columnspan=3, sticky="nsew"
        )
        # Perm Test
        perm_checkbox = gaita_widgets.checkbox(
            mainwindow,
            "Cluster-extent permutation test",
            cfg["do_permtest"],
            WIDGET_CFG,
        )
        perm_checkbox.grid(row=last_group_row + 6, column=0, columnspan=3, pady=10)
        # ANOVA info
        ANOVA_checkbox = gaita_widgets.checkbox(
            mainwindow,
            "ANOVA & Tukey's. If yes: choose factor-type below",
            cfg["do_anova"],
            WIDGET_CFG,
        )
        ANOVA_checkbox.configure(
            command=lambda: change_ANOVA_buttons_state(ANOVA_buttons),
        )  # dont use gui_utils here because its a bit special
        ANOVA_checkbox.grid(row=last_group_row + 7, column=0, columnspan=3, pady=10)
        # ANOVA design
        ANOVA_string_and_var_value = [
            ["Within-subjects (e.g. pre- & post-treatment)", "RM ANOVA"],
            ["Between-subjects (e.g. age)", "Mixed ANOVA"],
        ]
        ANOVA_buttons = []
        for i in range(len(ANOVA_string_and_var_value)):
            ANOVA_buttons.append(
                ctk.CTkRadioButton(
                    mainwindow,
                    text=ANOVA_string_and_var_value[i][0],
                    variable=cfg["anova_design"],
                    value=ANOVA_string_and_var_value[i][1],
                    fg_color=FG_COLOR,
                    hover_color=HOVER_COLOR,
                    font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
                )
            )
            if cfg["do_anova"].get() is True:
                ANOVA_buttons[i].configure(state="normal")
            ANOVA_buttons[-1].grid(row=last_group_row + 8 + i, column=0, columnspan=3)
        # initialise ANOVA buttons state by running this function once
        change_ANOVA_buttons_state(ANOVA_buttons)

        # ....................  advanced cfg & define features  ........................
        # empty label 3 for spacing
        empty_label_three = ctk.CTkLabel(mainwindow, text="")
        empty_label_three.grid(
            row=last_group_row + 10, column=0, columnspan=3, sticky="nsew"
        )

        # advanced cfg
        cfgwindow_button = gaita_widgets.header_button(
            mainwindow, "Advanced Configuration", WIDGET_CFG
        )
        cfgwindow_button.configure(
            command=lambda: (advanced_cfgwindow(mainwindow)),
        )
        cfgwindow_button.grid(
            row=last_group_row + 11,
            column=0,
            columnspan=3,
        )

        # define features button
        definefeatures_button = gaita_widgets.header_button(
            mainwindow, "I am ready - define features!", WIDGET_CFG
        )
        definefeatures_button.configure(
            command=lambda: (
                definefeatures_window(
                    mainwindow,
                    group_names,
                    group_dirs,
                    results_dir,
                    load_dir,  # tk var
                    cfg["which_leg"],  # tk var
                    root,
                ),
            ),
        )
        definefeatures_button.grid(row=last_group_row + 12, column=0, columnspan=3)

        # empty label four for spacing
        empty_label_four = ctk.CTkLabel(mainwindow, text="")
        empty_label_four.grid(
            row=last_group_row + 13, column=1, columnspan=3, sticky="ns"
        )

        # exit button
        exit_button = gaita_widgets.exit_button(mainwindow, WIDGET_CFG)
        exit_button.configure(
            command=lambda: (
                mainwindow.withdraw(),
                root.deiconify(),
                mainwindow.after(5000, mainwindow.destroy),
            ),
        )
        exit_button.grid(row=last_group_row + 14, column=0, columnspan=3)

        # maximise widgets to fit fullscreen
        maximise_widgets(mainwindow)


# %%..............  LOCAL FUNCTION(S) #2 - BUILD ADVANCED CFG WINDOW  ..................


def advanced_cfgwindow(mainwindow):
    """Build advanced configuration window"""

    # build window
    cfgwindow = ctk.CTkToplevel(mainwindow)
    cfgwindow.title("Advanced Configuration")
    cfgwindow.geometry(
        f"{int(screen_width / 2)}x{screen_height}+{int(screen_width / 4)}+0"
    )
    fix_window_after_its_creation(cfgwindow)

    # number of permutations
    permutation_number_label, permutation_number_entry = (
        gaita_widgets.label_and_entry_pair(
            cfgwindow,
            "Number of permutations of the cluster-extent test",
            cfg["permutation_number"],
            WIDGET_CFG,
        )
    )
    permutation_number_label.grid(row=0, column=0)
    permutation_number_entry.grid(row=1, column=0, sticky="n")

    # statistical threshold of significance
    stats_threshold_label, stats_threshold_entry = gaita_widgets.label_and_entry_pair(
        cfgwindow,
        "Alpha level of statistical significance (as a decimal)",
        cfg["stats_threshold"],
        WIDGET_CFG,
    )
    stats_threshold_label.grid(row=2, column=0)
    stats_threshold_entry.grid(row=3, column=0, sticky="n")

    # number of PCs
    PCA_n_components_string = (
        "Number of principal components (PCs). 0<PCs<1 for var-explained approach."
    )
    PCA_n_components_label, PCA_n_components_entry = gaita_widgets.label_and_entry_pair(
        cfgwindow,
        PCA_n_components_string,
        cfg["PCA_n_components"],
        WIDGET_CFG,
    )
    PCA_n_components_label.grid(row=4, column=0)
    PCA_n_components_entry.grid(row=5, column=0, sticky="n")

    # custom scatter info
    custom_scatter_label, custom_scatter_entry = gaita_widgets.label_and_entry_pair(
        cfgwindow,
        "Enter PCs for custom scatterplots (e.g. 2,3,5;4,6;3,4,5)",
        cfg["PCA_custom_scatter_PCs"],
        WIDGET_CFG,
    )
    custom_scatter_label.grid(row=6, column=0)
    custom_scatter_entry.grid(row=7, column=0, sticky="n")

    # PCA bins
    PCA_bins_label, PCA_bins_entry = gaita_widgets.label_and_entry_pair(
        cfgwindow,
        "Restrict PCA features to following cycle percentages (e.g. 0-10, 24, 50-75)",
        cfg["PCA_bins"],
        WIDGET_CFG,
    )
    PCA_bins_label.grid(row=8, column=0)
    PCA_bins_entry.grid(row=9, column=0, sticky="n")

    # save 3D PCA video
    save_PCA_video_checkbox = gaita_widgets.checkbox(
        cfgwindow,
        "Save video of 3D PCA Scatterplot (requires ffmpeg!)",
        cfg["PCA_save_3D_video"],
        WIDGET_CFG,
    )
    save_PCA_video_checkbox.grid(row=10, column=0)

    # color palette
    color_palette_string = "Choose figures' color palette"
    color_palette_label = ctk.CTkLabel(
        cfgwindow, text=color_palette_string, font=(TEXT_FONT_NAME, TEXT_FONT_SIZE)
    )
    color_palette_label.grid(row=11, column=0)
    color_palette_entry = ctk.CTkOptionMenu(
        cfgwindow,
        values=COLOR_PALETTES_LIST,
        variable=cfg["color_palette"],
        fg_color=FG_COLOR,
        button_color=FG_COLOR,
        button_hover_color=HOVER_COLOR,
        font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
    )
    color_palette_entry.grid(row=12, column=0, sticky="n")

    # plot SE
    plot_SE_box = gaita_widgets.checkbox(
        cfgwindow,
        "Plot standard error instead of standard deviation as error bars",
        cfg["plot_SE"],
        WIDGET_CFG,
    )
    plot_SE_box.grid(row=13, column=0)

    # legend outside
    legend_outside_checkbox = gaita_widgets.checkbox(
        cfgwindow,
        "Plot legends outside of figures' panels",
        cfg["legend_outside"],
        WIDGET_CFG,
    )
    legend_outside_checkbox.grid(row=14, column=0)

    # dont show plots
    dont_show_plots_checkbox = gaita_widgets.checkbox(
        cfgwindow,
        "Don't show plots in GUI (save only)",
        cfg["dont_show_plots"],
        WIDGET_CFG,
    )
    dont_show_plots_checkbox.grid(row=15, column=0)

    # which leg of human data to analyse
    which_leg_string = (
        "If first-level was AutoGaitA Universal 3D: which leg's step-cycles to analyse?"
    )
    which_leg_label = ctk.CTkLabel(
        cfgwindow, text=which_leg_string, font=(TEXT_FONT_NAME, TEXT_FONT_SIZE)
    )
    which_leg_label.grid(row=16, column=0)
    which_leg_options = ["left", "right"]  # !!! NU - "both" functionality
    which_leg_optionmenu = ctk.CTkOptionMenu(
        cfgwindow,
        values=which_leg_options,
        variable=cfg["which_leg"],
        fg_color=FG_COLOR,
        button_color=FG_COLOR,
        button_hover_color=HOVER_COLOR,
        font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
    )
    which_leg_optionmenu.grid(row=17, column=0, sticky="n")

    # done button
    adv_cfg_done_button = ctk.CTkButton(
        cfgwindow,
        text="I am done, update cfg.",
        fg_color=FG_COLOR,
        hover_color=HOVER_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
        command=lambda: cfgwindow.destroy(),
    )
    adv_cfg_done_button.grid(row=18, column=0, sticky="nsew", pady=20, padx=80)

    # maximise widgets to fit fullscreen
    maximise_widgets(cfgwindow)


# %%..............  LOCAL FUNCTION(S) #3 - BUILD ADD FEATURES WINDOW  ..................


def definefeatures_window(
    mainwindow, group_names, group_dirs, results_dir, load_dir, which_leg, root
):
    """Build define features window"""

    # nested function (called by run-button): extract boolean checkbox vars and store
    # in cfg dicts!
    def get_selected_variables():
        for key in LIST_VARS:
            # this list comprehension extracts all strings (keys) of checkbox_vars'
            # given sub-dict (either stats_variables or PCA_variables) and stores it to
            # the correct key of our cfg dict if it was checked (i.e., if var.get() is
            # True!)
            cfg[key] = [
                string for string, var in checkbox_vars[key].items() if var.get()
            ]

    # ...................  EXTRACT FEATURES FROM AN AVERAGE SC XLS  ....................
    # => Only needed if not loading a previous run (note there is a check for features
    #    being present in all dfs in load_previous_runs_dataframes in group_main)
    # => First read the xls as df, extract columns that are meaningful after
    #    averaging, or throw errors if we don't manage to do so
    #    -- (note we extract from average xls since that automatically informs about
    #        export_average_x/y vars!)
    if len(load_dir.get()) > 0:
        group_one_string = group_names[0].get()
        load_dir_string = load_dir.get()
        df = pd.read_excel(
            os.path.join(
                load_dir_string, f"{group_one_string} - {AVG_GROUP_SHEET_NAME}.xlsx"
            )
        )
    else:
        df = None
        test_directories = []  # first see if all group dirs are valid paths
        for directory in group_dirs:
            test_directories.append(directory.get())
        for directory in test_directories:
            if not os.path.exists(directory):
                error_msg = "No directory found at: " + directory
                tk.messagebox.showerror(title="Folder not found!", message=error_msg)
                return
        if not results_dir.get():
            error_msg = "You did not specify a directory to save results to!"
            tk.messagebox.showerror(title="Folder not found!", message=error_msg)
            return
        if not os.path.exists(results_dir.get()):
            os.makedirs(results_dir.get())
        some_groups_dir = group_dirs[0].get()
        all_ID_dirs = [
            f
            for f in os.listdir(some_groups_dir)
            if os.path.isdir(os.path.join(some_groups_dir, f))
        ]
        for ID_dir in all_ID_dirs:
            some_IDs_dir = os.path.join(some_groups_dir, ID_dir)
            IDs_files = os.listdir(some_IDs_dir)
            # next operator means we loop lazy - stop once we find it & return None if av
            # sheet not in dir
            av_sheet_path = next(
                (file for file in IDs_files if NORM_SHEET_NAME in file), None
            )
            if av_sheet_path:  # won't be true if no AVXLS found
                full_path = os.path.join(some_IDs_dir, av_sheet_path)
                if av_sheet_path.endswith(".xlsx"):
                    try:
                        # universal 3D (that has which_leg sheets) always exports xlsx
                        df = pd.read_excel(full_path, sheet_name=which_leg.get())
                    except:
                        df = pd.read_excel(full_path)
                    break
                elif av_sheet_path.endswith(".csv"):
                    df = pd.read_csv(full_path)
                    break
    # select columns we want to provide as feature options using regular expressions
    # => since it's on av sheet, x & Y will only be included if export_average_x/Y was
    #    True @ first-level
    if df is not None:
        feature_strings = tuple(
            df.filter(regex="(x$|y$|Y$|Z$|Angle$|Velocity$|Acceleration$)").columns
        )
    else:
        error_string = "Unable to find any Average SC sheet at " + some_groups_dir
        tk.messagebox.showerror(title="Group Directory Error!", message=error_string)
        print(error_string)
        print("Analysis did not run due to above errors. Fix them & re-try!")
        return

    # .......................  BUILD THE TOPLEVEL WINDOW  .............................
    # build window
    featureswindow = ctk.CTkToplevel(mainwindow)
    featureswindow.title("Add Features & Run!")
    featureswindow.geometry(f"{screen_width}x{screen_height}+0+0")
    fix_window_after_its_creation(featureswindow)
    # grid prep
    scrollbar_rows = 5  # scrollbar rowspan (in featureswindow)
    scrollbar_grid_ncols = 3
    # + 5 to be sure we have enough rows
    scrollbar_grid_nrows = math.ceil(len(feature_strings) / scrollbar_grid_ncols) + 5

    # stats label
    stats_label = gaita_widgets.header_label(featureswindow, "Statistics", WIDGET_CFG)
    stats_label.grid(row=0, column=0, sticky="nsew")
    # PCA label
    PCA_label = gaita_widgets.header_label(featureswindow, "PCA", WIDGET_CFG)
    PCA_label.grid(row=0, column=1, sticky="nsew")

    # .........................  PCA / STATS FEATURE FRAMES  ...........................
    # stats scrollable frame initialisation
    stats_frame = ctk.CTkScrollableFrame(featureswindow)
    stats_frame.grid(row=1, column=0, rowspan=scrollbar_rows, sticky="nsew")
    # PCA scrollable frame initialisation
    PCA_frame = ctk.CTkScrollableFrame(featureswindow)
    PCA_frame.grid(row=1, column=1, rowspan=scrollbar_rows, sticky="nsew")
    # populate scrollable frames with features
    # => note that checkbox_vars just store whether a given checkbox is checked
    #    (i.e., True) or not. we will store all true checkboxes for both PCA and stats
    #    later (but it's handy that the key names overlap)
    checkbox_vars = {key: {} for key in LIST_VARS}
    for frame in [stats_frame, PCA_frame]:
        if frame == stats_frame:  # make sure to put stuff into the correct dict
            key = "stats_variables"
        elif frame == PCA_frame:
            key = "PCA_variables"
        row_counter = 1  # index rows at 1 for the top otherwise it breaks (dont ask)
        col_counter = 0
        for feature in feature_strings:
            # check whether features of this run (feature in feature_strings) where
            # also in the last run (cfg[key] = stats/PCA_variables list from config file)
            # if yes then set the checkbox to true
            if feature in cfg["last_runs_" + key]:
                checkbox_vars[key][feature] = var = tk.BooleanVar(value=True)
            # else set it to false as by default
            else:
                checkbox_vars[key][feature] = var = tk.BooleanVar()
            this_checkbox = gaita_widgets.checkbox(frame, feature, var, WIDGET_CFG)
            this_checkbox.grid(row=row_counter, column=col_counter, sticky="nsew")
            # if user doesn't want to do stats, dont have them choose features
            if frame == stats_frame:
                if (cfg["do_permtest"].get() is False) & (
                    cfg["do_anova"].get() is False
                ):
                    this_checkbox.configure(state="disabled")
            row_counter += 1
            # # CARE - the -1 is important here since indexing starts at 0 and grid_nrows
            # # is like len() kinda
            if row_counter == scrollbar_grid_nrows:  # not nrows-1 because we start @ 1
                row_counter = 1  # index rows at 1 for the top otherwise it breaks
                col_counter += 1
        # maximise widgets to fit frame (of scrollbar!)
        maximise_widgets(frame)

    # run
    run_button = gaita_widgets.header_button(
        featureswindow, "I am ready - run analysis!", WIDGET_CFG
    )
    run_button.configure(
        command=lambda: (
            build_donewindow(
                group_names,
                group_dirs,
                results_dir,
                load_dir,
                root,
                mainwindow,
                featureswindow,
            ),
            get_selected_variables(),
        ),
    )
    run_button.grid(
        row=1 + scrollbar_rows, column=0, columnspan=2, sticky="nsew", pady=20, padx=200
    )

    # maximise widgets to fit fullscreen
    maximise_widgets(featureswindow)


# %%..................  LOCAL FUNCTION(S) #4 - BUILD DONE WINDOW  ......................
def build_donewindow(
    group_names, group_dirs, results_dir, load_dir, root, mainwindow, featureswindow
):
    """Build done window"""
    # create done window & make it pretty and nice
    w = screen_width
    h = screen_height
    donewindow = ctk.CTkToplevel(root)
    donewindow.title("GaitA Ready :)")
    donewindow_w = w * (3 / 5)
    donewindow_h = h * (1 / 5)
    donewindow_x = (w - donewindow_w) // 2
    donewindow_y = h * (1 / 2.5)
    donewindow.geometry(
        "%dx%d+%d+%d" % (donewindow_w, donewindow_h, donewindow_x, donewindow_y)
    )
    fix_window_after_its_creation(donewindow)

    # labels
    done_label1_string = "Your results will be saved as your data is processed"
    done_label1 = ctk.CTkLabel(
        donewindow,
        text=done_label1_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    done_label1.grid(row=0, column=0, sticky="nsew")
    done_label2_string = (
        "Please see the Python command window for progress "
        + "and the figure panel for an overview of all plots."
    )
    done_label2 = ctk.CTkLabel(
        donewindow,
        text=done_label2_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    done_label2.grid(row=1, column=0, sticky="nsew")
    done_label3_string = (
        "You may start another analysis while we are "
        + "processing - however your PC might slow down a bit. "
    )
    done_label3 = ctk.CTkLabel(
        donewindow,
        text=done_label3_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    done_label3.grid(row=2, column=0, sticky="nsew")
    done_label4_string = (
        "Thank you for using AutoGaitA! Feel free to "
        + "contact me on Github or at autogaita@fz-juelich.de - MH."
    )
    done_label4 = ctk.CTkLabel(
        donewindow,
        text=done_label4_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    done_label4.grid(row=3, column=0, sticky="nsew")

    # ..................................................................................
    # ........................  IMPORTANT - prepare folderinfo .........................
    # ..................................................................................
    folderinfo = {
        "group_names": [],
        "group_dirs": [],
        "results_dir": "",
        "load_dir": "",
    }
    for g in range(len(group_names)):
        folderinfo["group_names"].append(group_names[g].get())
        folderinfo["group_dirs"].append(group_dirs[g].get())
    folderinfo["results_dir"] = results_dir.get()
    folderinfo["load_dir"] = load_dir.get()

    # run button
    done_button = ctk.CTkButton(
        donewindow,
        text="Run!",
        fg_color=FG_COLOR,
        hover_color=HOVER_COLOR,
        font=(HEADER_FONT_NAME, MAIN_HEADER_FONT_SIZE),
        command=lambda: (
            update_config_file(folderinfo, cfg),
            run_analysis(folderinfo, cfg),
            mainwindow.destroy(),
            featureswindow.destroy(),
            donewindow.destroy(),
            root.deiconify(),
        ),
    )
    done_button.grid(row=4, column=0, sticky="nsew", pady=10, padx=200)

    maximise_widgets(donewindow)


# %%..................  LOCAL FUNCTION(S) #5 - RUN GROUP ANALYSIS  .....................
def run_analysis(folderinfo, cfg):
    """Run group results based on current info

    Note
    ----
    1) We previously used .get() on folderinfo string entries (names & dirs)
    => see IMPORTANT - prepare folderinfo above
    2) check_folderinfo_and_cfg just checks user inputs for validity
    => no .get() is called there, no variables are assigned to anything
    => tbh I think the check for dirs is redundant since we already do that at start
       of definefeatures_window - but whatever
    3) extract_this_runs_folderinfo_and_cfg extracts vars using .get() if necessary
    => we return this_ folderinfo/cfg dicts which are then used for running the analysis
    => ie. folderinfo & cfg entries can be changed by user, their values (eg. entries &
       checkboxes) are only collected once the user hits the run button in the features
       window (run_button collects stats/PCA checkboxes & done_button everything else)
    """

    # check if user-provided values were correct before extracting them for run function
    error_msg = check_folderinfo_and_cfg(folderinfo, cfg)
    if len(error_msg) > 0:
        error_string = "The following input variables are wrong! Fix them & re-run!\n"
        for i in range(len(error_msg)):
            error_string = error_string + "\n" + error_msg[i]
        tk.messagebox.showerror(title="Input variable errors!", message=error_string)
        print(error_string)
        print("Analysis did not run due to above errors. Fix them & re-try!")
        return

    # extract this runs variables correctly before running
    this_runs_folderinfo, this_runs_cfg = extract_this_runs_folderinfo_and_cfg(
        folderinfo, cfg
    )

    # run in a thread (see a note about thread running after helper functions)
    run_thread = Thread(
        target=autogaita.group, args=(this_runs_folderinfo, this_runs_cfg)
    )
    run_thread.start()


def check_folderinfo_and_cfg(folderinfo, cfg):
    """Check if the folderinfo and cfg values provided by the user are fine"""
    error_msg = []
    both_dicts = [folderinfo, cfg]
    for inner_dict in both_dicts:
        for key in inner_dict.keys():
            # check string vars: group dirs & names and results dir
            if key in STRING_VARS:
                if key == "group_dirs" and len(folderinfo["load_dir"]) > 0:
                    for g_idx, group_dir in enumerate(inner_dict[key]):
                        if not os.path.exists(group_dir):
                            this_msg = (
                                "Group #"
                                + str(g_idx + 1)
                                + "'s directory does not exist!"
                            )
                            print("\n" + this_msg)
                            error_msg.append(this_msg)
                elif key == "group_names":
                    for i in range(len(inner_dict[key])):
                        if len(inner_dict[key][i]) == 0:
                            this_msg = "Group #" + str(i + 1) + "'s name is empty!"
                            print("\n" + this_msg)
                            error_msg.append(this_msg)
                elif key == "results_dir":
                    if len(inner_dict[key]) == 0:
                        this_msg = key + " is empty!"
                        print("\n" + this_msg)
                        error_msg.append(this_msg)
            # check if numerical cfg variables were given correctly
            if key in FLOAT_VARS:
                try:
                    float(inner_dict[key].get())
                except ValueError:
                    this_msg = (
                        key
                        + " value was not a decimal number, but "
                        + inner_dict[key].get()
                    )
                    print(this_msg)
                    error_msg.append(this_msg)
            if key in INT_VARS:
                try:
                    int(inner_dict[key].get())
                except:
                    this_msg = (
                        key
                        + " value was not an integer number, but "
                        + inner_dict[key].get()
                    )
                    print(this_msg)
                    error_msg.append(this_msg)
    return error_msg


def extract_this_runs_folderinfo_and_cfg(folderinfo, cfg):
    """Extract folderinfo and cfg of current run - immediately before running!"""
    # prepare this runs folderinfo
    this_runs_folderinfo = {}
    # make sure that directories end with os.sep (since we want to be able to save files
    # using string concatenation). If not just add a forward slash. In the beginning
    # of autogaita.group, we will convert os.sep to forward slashes if they should be
    # backward slashes (windows works with both)
    # => The part that does this in autogaita.group was updated and uses os.path.join
    # => Thought about changing strings to Paths here but will leave it for now it works
    for key in folderinfo.keys():
        if key == "group_names":
            this_runs_folderinfo[key] = folderinfo[key]
        if "dir" in key:
            if isinstance(folderinfo[key], list):  # group_dirs
                this_runs_folderinfo[key] = [[] for _ in range(len(folderinfo[key]))]
                for i in range(len(folderinfo[key])):
                    if not folderinfo[key][i].endswith(os.sep):
                        this_runs_folderinfo[key][i] = folderinfo[key][i] + "/"
                    else:
                        this_runs_folderinfo[key][i] = folderinfo[key][i]
            # results_dir and load_dir (check load_dir only if not empty)
            elif key in ["results_dir", "load_dir"]:
                if len(folderinfo[key]) > 0 and not folderinfo[key].endswith(os.sep):
                    this_runs_folderinfo[key] = folderinfo[key] + "/"
                else:
                    this_runs_folderinfo[key] = folderinfo[key]
    # prepare this runs cfg
    this_runs_cfg = {}
    for key in cfg.keys():
        if key in LIST_VARS:  # no get() needed - since these are lists already
            this_runs_cfg[key] = cfg[key]
            # elif is necessary as we want to exclude "last_runs_stats/PCA_variables"
        elif key not in EXCLUDED_VARS_FROM_CFG_FILE:
            this_runs_cfg[key] = cfg[key].get()
            # convert to numbers
            if key in FLOAT_VARS:
                this_runs_cfg[key] = float(this_runs_cfg[key])
            if key in INT_VARS:
                this_runs_cfg[key] = int(this_runs_cfg[key])
    return this_runs_folderinfo, this_runs_cfg

    # !!! NU - leave as is for now but if you want to add info messages for users about
    # how many active runs they have in their pipeline when adding a new one you can probably
    # have run_thread be a global variable and then use len(globals["run_thread"].is_active())
    # or something to print the number of active threads

    # # once we got this done, run the analysis in a separate thread
    # local_vars = locals()
    # if "run_thread" in local_vars:
    #     if local_vars["run_thread"].is_alive():
    #         still_busy_message = (
    #             "Your previous run is still being processed, we will let you know "
    #             + "once we are ready for another run :)"
    #             )
    #         tk.messagebox(title="Still busy :s", message=still_busy_message)
    #     else:
    #         run_thread = Thread(
    #             target=autogaita_group.main, args=(this_runs_folderinfo, this_runs_cfg)
    #             )
    #         check_thread = Thread(target=check_run_thread_completion, args=(run_thread))
    #         run_thread.start()
    #         check_thread.start()
    # else:
    #     run_thread = Thread(
    #         target=autogaita_group_dev.main, args=(this_runs_folderinfo, this_runs_cfg)
    #         )
    #     check_thread = Thread(target=check_run_thread_completion, args=(run_thread,))
    #     run_thread.start()
    #     check_thread.start()

    # def check_run_thread_completion(run_thread):
    #     """Keep checking if run thread is finished, if so display a tkinter messagebox"""
    #     dummy_number = 1
    #     while dummy_number != 0:
    #         if run_thread.is_alive():
    #             dummy_number += 1
    #     else:
    #         done_message = (
    #             "Your run has finished execution! You can find your results in the "
    #             + "group directory you provided and are able to run further runs if "
    #             + "desired.")
    #         tk.messagebox(title="Finished :)", message=done_message)
    #         dummy_number = 0  # so we exit the while loop


# %%...............  LOCAL FUNCTION(S) #6 - VARIOUS HELPER FUNCTIONS  ..................


def change_ANOVA_buttons_state(ANOVA_buttons):
    """Change the state of ANOVA radio button widgets based on whether user wants
    to perform an ANOVA or not.
    """
    if cfg["do_anova"].get() is True:
        for i in range(len(ANOVA_buttons)):
            ANOVA_buttons[i].configure(state="normal")
    elif cfg["do_anova"].get() is False:
        for i in range(len(ANOVA_buttons)):
            ANOVA_buttons[i].configure(state="disabled")


def maximise_widgets(window):
    """Maximises all widgets to look good in fullscreen"""
    # fix the grid to fill the window
    num_rows = window.grid_size()[1]  # maximise rows
    for r in range(num_rows):
        window.grid_rowconfigure(r, weight=1)
    num_cols = window.grid_size()[0]  # maximise cols
    for c in range(num_cols):
        window.grid_columnconfigure(c, weight=1)


def fix_window_after_its_creation(window):
    """Perform some quality of life things after creating a window (root or Toplevel)"""
    window.attributes("-topmost", True)
    window.focus_set()
    window.after(100, lambda: window.attributes("-topmost", False))  # 100 ms


def update_config_file(folderinfo, cfg):
    """updates the group_gui_config file with this folderinfo and cfg parameters"""
    # transform tkVars into normal strings and bools
    output_dicts = [
        {"group_names": [], "group_dirs": [], "results_dir": "", "load_dir": ""},
        {},
    ]

    for i in range(len(output_dicts)):
        if i == 0:  # as the list index 0 refers to  folderinfo
            # to not alter the initial folderinfo variable a deepcopy is used
            input_dict = copy.deepcopy(folderinfo)
            output_dicts[0] = input_dict
            # the next if statement ensures that the names and dirs list always contains
            # the max group number of entries
            if len(output_dicts[0]["group_names"]) != MAX_GROUP_NUM:
                for i in range(MAX_GROUP_NUM - len(input_dict["group_names"])):
                    output_dicts[0]["group_names"].append("")
                    output_dicts[0]["group_dirs"].append("")

        elif i == 1:  # as the list index 1 refers to cfg
            input_dict = cfg
            for key in input_dict.keys():
                if key in LIST_VARS:  # PCA or Stats variables
                    # if list of strings, initialise output empty list and append vals
                    output_dicts[i][key] = []
                    for entry in input_dict[key]:
                        output_dicts[i][key].append(entry)
                # otherwise (if str, int or bool) get() and define
                elif (
                    key not in EXCLUDED_VARS_FROM_CFG_FILE
                ):  # not PCA or Stats variables
                    output_dicts[i][key] = input_dict[key].get()

    # merge the two configuration dictionaries
    # 0 = folderinfo, 1 = cfg, see above
    configs_list = [
        output_dicts[0],
        output_dicts[1],
    ]
    # write the configuration file
    with open(
        os.path.join(AUTOGAITA_FOLDER_PATH, CONFIG_FILE_NAME), "w"
    ) as config_json_file:
        json.dump(configs_list, config_json_file, indent=4)


def extract_cfg_from_json_file(root):
    """loads the cfg dictionary from the config file"""
    # load the configuration file
    with open(
        os.path.join(AUTOGAITA_FOLDER_PATH, CONFIG_FILE_NAME), "r"
    ) as config_json_file:
        # config_json contains list with 0 -> result and 1 -> cfg data
        last_runs_cfg = json.load(config_json_file)[1]

    cfg = {}
    # assign values to the cfg dict
    for key in last_runs_cfg.keys():
        if key in TK_BOOL_VARS:
            cfg[key] = tk.BooleanVar(root, last_runs_cfg[key])
        elif key in TK_STR_VARS:
            cfg[key] = tk.StringVar(root, last_runs_cfg[key])
        # PCA/stats_variable are not needed as tkStringVars
        # see the ---PCA / STATS FEATURE FRAMES--- section for their usage
        elif key in LIST_VARS:
            # creates an empty list for this runs variables
            cfg[key] = []
            # and saves the stats/PCA_variables from the last run in a separate list
            # note, this list will not be save to the config file as "last_runs_..._variables"
            # in not in LIST_VARS
            cfg["last_runs_" + key] = []
            for entry in last_runs_cfg[key]:
                cfg["last_runs_" + key].append(entry)

    return cfg


def extract_group_names_and_dirs_from_json_file():
    """loads the group names and dirs from the config file"""
    # load the configuration file
    with open(
        os.path.join(AUTOGAITA_FOLDER_PATH, CONFIG_FILE_NAME), "r"
    ) as config_json_file:
        # config_json contains list with 0 -> folderinfo and 1 -> cfg data
        last_runs_folderinfo = json.load(config_json_file)[0]

    group_names = []
    group_dirs = []

    for key in last_runs_folderinfo.keys():
        if key == "group_names":
            for entry in last_runs_folderinfo[key]:
                group_names.append(entry)
        if key == "group_dirs":
            for entry in last_runs_folderinfo[key]:
                group_dirs.append(entry)

    return group_names, group_dirs


def extract_results_and_load_dirs_from_json_files():
    """loads the results dir from the config file"""
    # load the configuration file
    with open(
        os.path.join(AUTOGAITA_FOLDER_PATH, CONFIG_FILE_NAME), "r"
    ) as config_json_file:
        # config_json contains list with 0 -> folderinfo and 1 -> cfg data
        last_runs_folderinfo = json.load(config_json_file)[0]

    results_dir = last_runs_folderinfo["results_dir"]
    load_dir = last_runs_folderinfo["load_dir"]

    return results_dir, load_dir


# %% what happens if we hit run
if __name__ == "__main__":
    run_group_gui()



# %%..............  SHARED 2D WINDOW - COLUMN & ANGLE INFO  .......................
# Note
# ----
# Hindlimb and forelimb joints were re-named to primary and secondary joints @ v0.4.0.
# I kept the old variable/key-names for compatibility with autogaita_dlc.
# => Except for input to build_beam_jointadd_window since that is used for a string
import autogaita.gui.gui_utils as gui_utils
import autogaita.gui.gaita_widgets as gaita_widgets
import customtkinter as ctk
import tkinter as tk


def build_column_info_window(root, cfg, widget_cfg, root_dimensions):
    """Build a window allowing users to configure custom column names if required"""
    columnwindow = ctk.CTkToplevel(root)
    columnwindow.geometry("%dx%d+%d+%d" % root_dimensions)
    columnwindow.title("Custom column names & features")
    gui_utils.fix_window_after_its_creation(columnwindow)

    # unpack
    FG_COLOR = widget_cfg["FG_COLOR"]
    HEADER_TXT_COLOR = widget_cfg["HEADER_TXT_COLOR"]
    HEADER_FONT_NAME = widget_cfg["HEADER_FONT_NAME"]
    HEADER_FONT_SIZE = widget_cfg["HEADER_FONT_SIZE"]
    CLOSE_COLOR = widget_cfg["CLOSE_COLOR"]
    CLOSE_HOVER_COLOR = widget_cfg["CLOSE_HOVER_COLOR"]

    # .............  Nested Function: Add joint label & entry  .........................
    def add_joint(cfg, window, key):
        """Add a joint if needed or a angle name/lower_joint/upper_joint if angle widget"""
        # append StringVar to cfg appropriately
        if key == "angles":
            for angle_key in cfg[key]:
                cfg[key][angle_key].append(tk.StringVar(root, ""))
        else:
            cfg[key].append(tk.StringVar(root, ""))
        # find out the number of rows to append to it
        nrows = window.grid_size()[1]
        # add stuff based on to which case we are adding
        if key in [
            "hind_joints",
            "fore_joints",
            "beam_hind_jointadd",
            "beam_fore_jointadd",
        ]:
            if key == "hind_joints":
                label_string = "Primary Joint #" + str(len(cfg[key]))
            elif key == "fore_joints":
                label_string = "Secondary Joint #" + str(len(cfg[key]))
            elif key == "beam_hind_jointadd":
                label_string = "Left beam-subtraction joint #" + str(len(cfg[key]))
            elif key == "beam_fore_jointadd":
                label_string = "Right beam-subtraction joint #" + str(len(cfg[key]))
            label, entry = gaita_widgets.label_and_entry_pair(
                window, label_string, cfg[key][-1], widget_cfg
            )
            label.grid(row=nrows + 1, column=0, sticky="ew")
            entry.grid(row=nrows + 2, column=0)
        elif key == "angles":
            for a, angle_key in enumerate(cfg[key]):
                if angle_key == "name":
                    this_case = "Angle"
                elif angle_key == "lower_joint":
                    this_case = "Lower Joint"
                elif angle_key == "upper_joint":
                    this_case = "Upper Joint"
                label, entry = gaita_widgets.label_and_entry_pair(
                    window,
                    this_case + " #" + str(len(cfg[key][angle_key])),
                    cfg[key][angle_key][-1],
                    widget_cfg,
                )
                label.grid(row=nrows + 1, column=angle_column + a, sticky="ew")
                entry.grid(row=nrows + 2, column=angle_column + a)
        # maximise columns
        for c in range(window.grid_size()[0]):
            window.grid_columnconfigure(c, weight=1)

    # ...............  Nested Function: Beam jointadd window  ..........................
    def build_beam_window(cfg, widget_cfg):
        """Build a window for configuring the beam, i.e.:
        1. beam_col_left and right
        ==> what were the col names of your beam in the beam file
        2. beam_hind_jointadd and beam_fore_jointadd
        ==> what additional joints (not included in hind/fore) do you want to include in beam standardisation?
        """

        # build window and fullscreen
        beamwindow = ctk.CTkToplevel(columnwindow)
        beamwindow.geometry("%dx%d+%d+%d" % (root_dimensions))
        beamwindow.title("Beam Configuration")
        gui_utils.fix_window_after_its_creation(beamwindow)
        beam_scrollable_rows = 1
        total_padx = 20
        left_padx = (total_padx, total_padx / 2)
        right_padx = (total_padx / 2, total_padx)
        # ................... left section - left beam / hind joints ...................
        # left beam label & entry
        beam_left_label, beam_left_entry = gaita_widgets.label_and_entry_pair(
            beamwindow,
            "Left beam column (primary joints')",
            cfg["beam_col_left"][0],
            widget_cfg,
        )
        beam_left_label.grid(row=0, column=0, sticky="nsew")
        beam_left_entry.grid(row=1, column=0)
        # empty label one (for spacing)
        empty_label_one = ctk.CTkLabel(beamwindow, text="")
        empty_label_one.grid(row=2, column=0)
        # important: cfg key for forelimb joint add
        hindlimb_key = "beam_hind_jointadd"
        # hindlimb jointadd label
        hind_jointsubtract_label = ctk.CTkLabel(
            beamwindow,
            text="Left beam: additional joints subtracted",
            fg_color=FG_COLOR,
            text_color=HEADER_TXT_COLOR,
            font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
        )
        hind_jointsubtract_label.grid(row=3, column=0, sticky="nsew")
        # initialise scrollable frame for beamsubtract windows
        hind_jointsubtract_frame = ctk.CTkScrollableFrame(beamwindow)
        hind_jointsubtract_frame.grid(
            row=4,
            column=0,
            rowspan=beam_scrollable_rows,
            sticky="nsew",
        )
        # initialise labels & entries
        initialise_labels_and_entries(
            cfg,
            widget_cfg,
            hind_jointsubtract_frame,
            hindlimb_key,
            "Left beam-subtraction joint",
        )
        # add button
        add_hindjoint_button = gaita_widgets.header_button(
            beamwindow,
            "Add left beam-subtraction joint",
            widget_cfg,
        )
        add_hindjoint_button.configure(
            command=lambda: add_joint(
                cfg, hind_jointsubtract_frame, hindlimb_key
            ),  # input = cfg's key
        )
        add_hindjoint_button.grid(
            row=4 + beam_scrollable_rows,
            column=0,
            sticky="nsew",
            padx=left_padx,
            pady=20,
        )
        # .................. right section - right beam / fore joints ..................
        # right beam label & entry
        beam_right_label, beam_right_entry = gaita_widgets.label_and_entry_pair(
            beamwindow,
            "Right beam column (secondary joints')",
            cfg["beam_col_right"][0],
            widget_cfg,
        )
        beam_right_label.grid(row=0, column=1, sticky="nsew")
        beam_right_entry.grid(row=1, column=1)
        # empty label two (for spacing)
        empty_label_two = ctk.CTkLabel(beamwindow, text="")
        empty_label_two.grid(row=2, column=1)
        # important: cfg key for forelimb joint add
        forelimb_key = "beam_fore_jointadd"
        # hindlimb jointadd label
        fore_jointsubtract_label = ctk.CTkLabel(
            beamwindow,
            text="Right beam: additional joints subtracted",
            fg_color=FG_COLOR,
            text_color=HEADER_TXT_COLOR,
            font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
        )
        fore_jointsubtract_label.grid(row=3, column=1, sticky="nsew")
        # initialise scrollable frame for beamsubtract windows
        fore_jointsubtract_frame = ctk.CTkScrollableFrame(beamwindow)
        fore_jointsubtract_frame.grid(
            row=4,
            column=1,
            rowspan=beam_scrollable_rows,
            sticky="nsew",
        )
        # initialise labels & entries
        initialise_labels_and_entries(
            cfg,
            widget_cfg,
            fore_jointsubtract_frame,
            forelimb_key,
            "Right beam-subtraction joint",
        )
        # add button
        add_forejoint_button = gaita_widgets.header_button(
            beamwindow,
            "Add right beam-subtraction joint",
            widget_cfg,
        )
        add_forejoint_button.configure(
            command=lambda: add_joint(
                cfg, fore_jointsubtract_frame, forelimb_key
            ),  # input = cfg's key
        )
        add_forejoint_button.grid(
            row=4 + beam_scrollable_rows,
            column=1,
            sticky="nsew",
            padx=right_padx,
            pady=20,
        )
        # .................... bottom section - update & close  ........................
        # empty label three (for spacing)
        empty_label_three = ctk.CTkLabel(beamwindow, text="")
        empty_label_three.grid(row=5 + beam_scrollable_rows, column=0, columnspan=2)
        # done button
        beam_done_button = ctk.CTkButton(
            beamwindow,
            text="I am done, update cfg!",
            fg_color=CLOSE_COLOR,
            hover_color=CLOSE_HOVER_COLOR,
            font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
            command=lambda: beamwindow.destroy(),
        )
        beam_done_button.grid(
            row=6 + beam_scrollable_rows,
            column=0,
            columnspan=2,
            sticky="nsew",
            padx=20,
            pady=20,
        )
        # maximise widgets
        gui_utils.maximise_widgets(beamwindow)

    # ...................  Scrollable Window Configuration  ............................
    scrollable_rows = 7

    # ...................  Column 0: hind limb joint names  ............................
    hind_column = 0
    # header label
    hind_joint_label = ctk.CTkLabel(
        columnwindow,
        text="Primary Joints",
        fg_color=FG_COLOR,
        text_color=HEADER_TXT_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
    )
    hind_joint_label.grid(row=0, column=hind_column, sticky="nsew", pady=(0, 5))
    # initialise scrollable frame for hindlimb
    hindlimb_frame = ctk.CTkScrollableFrame(columnwindow)
    hindlimb_frame.grid(
        row=1,
        column=hind_column,
        rowspan=scrollable_rows,
        sticky="nsew",
    )
    # initialise labels & entries with hind limb defaults
    initialise_labels_and_entries(
        cfg, widget_cfg, hindlimb_frame, "hind_joints", "Primary Joint "
    )
    # add joint button
    add_hind_joint_button = gaita_widgets.header_button(
        columnwindow,
        "Add Primary Joint",
        widget_cfg,
    )
    add_hind_joint_button.configure(
        command=lambda: add_joint(
            cfg, hindlimb_frame, "hind_joints"
        ),  # 2nd input = cfg's key
    )
    add_hind_joint_button.grid(
        row=2 + scrollable_rows, column=hind_column, sticky="nsew", padx=5, pady=(10, 5)
    )
    # beam config window label
    beam_window_label = ctk.CTkLabel(columnwindow, text="")  # empty for spacing only
    beam_window_label.grid(
        row=4 + scrollable_rows,
        column=hind_column,
        columnspan=2,
        sticky="nsew",
        padx=1,
        pady=(0, 5),
    )
    # beam config window button
    beam_window_button = gaita_widgets.header_button(
        columnwindow,
        "Baseline (Beam) Configuration",
        widget_cfg,
    )
    beam_window_button.configure(
        command=lambda: build_beam_window(cfg, widget_cfg),
    )
    beam_window_button.grid(
        row=5 + scrollable_rows,
        column=hind_column,
        columnspan=2,
        sticky="nsew",
        padx=40,
        pady=(10, 5),
    )

    # ...................  Column 1: fore limb joint names  ............................
    fore_column = 1
    # header label
    fore_joint_label = ctk.CTkLabel(
        columnwindow,
        text="Secondary Joints",
        fg_color=FG_COLOR,
        text_color=HEADER_TXT_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
    )
    fore_joint_label.grid(row=0, column=fore_column, sticky="nsew", pady=(0, 5))
    # initialise scrollable frame for forelimb
    forelimb_frame = ctk.CTkScrollableFrame(columnwindow)
    forelimb_frame.grid(
        row=1, column=fore_column, rowspan=scrollable_rows, sticky="nsew"
    )
    # initialise labels & entries with fore limb defaults
    initialise_labels_and_entries(
        cfg,
        widget_cfg,
        forelimb_frame,
        "fore_joints",
        "Secondary Joint ",
    )
    # add joint button
    add_fore_joint_button = gaita_widgets.header_button(
        columnwindow,
        "Add Secondary Joint",
        widget_cfg,
    )
    add_fore_joint_button.configure(
        command=lambda: add_joint(
            cfg, forelimb_frame, "fore_joints"
        ),  # 2nd input = cfg's key
    )
    add_fore_joint_button.grid(
        row=2 + scrollable_rows, column=fore_column, sticky="nsew", padx=5, pady=(10, 5)
    )

    # .........  Column 2: angle names/joint-definitions & done button  ................
    angle_column = 2
    # header label
    angle_label = ctk.CTkLabel(
        columnwindow,
        text="Angle Configuration",
        fg_color=FG_COLOR,
        text_color=HEADER_TXT_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
    )
    angle_label.grid(
        row=0, column=angle_column, columnspan=3, sticky="nsew", pady=(0, 5)
    )
    # initialise scrollable frame for angles
    angle_frame = ctk.CTkScrollableFrame(columnwindow)
    angle_frame.grid(
        row=1, column=angle_column, rowspan=scrollable_rows, columnspan=3, sticky="nsew"
    )
    # initial entries & labels
    for a, angle_key in enumerate(cfg["angles"]):
        if angle_key == "name":
            this_case = "Angle"
        elif angle_key == "lower_joint":
            this_case = "Lower Joint"
        elif angle_key == "upper_joint":
            this_case = "Upper Joint"
        initialise_labels_and_entries(
            cfg,
            widget_cfg,
            angle_frame,
            ["angles", angle_key],
            this_case,
            angle_column + a,
        )
    # add angle trio button
    add_angle_button = gaita_widgets.header_button(
        columnwindow,
        "Add Angle",
        widget_cfg,
    )
    add_angle_button.configure(
        command=lambda: add_joint(cfg, angle_frame, "angles"),  # 2nd input = cfg's key
    )
    add_angle_button.grid(
        row=2 + scrollable_rows,
        column=angle_column,
        columnspan=3,
        sticky="nsew",
        padx=5,
        pady=(10, 5),
    )
    # done label
    done_label = ctk.CTkLabel(columnwindow, text="")  # empty for spacing only
    done_label.grid(
        row=4 + scrollable_rows,
        column=angle_column,
        columnspan=3,
        sticky="nsew",
        padx=1,
        pady=(0, 5),
    )
    # done button
    columncfg_done_button = ctk.CTkButton(
        columnwindow,
        text="I am done, update cfg!",
        fg_color=CLOSE_COLOR,
        hover_color=CLOSE_HOVER_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
        command=lambda: columnwindow.destroy(),
    )
    columncfg_done_button.grid(
        row=5 + scrollable_rows,
        column=angle_column,
        columnspan=3,
        sticky="nsew",
        padx=40,
        pady=(10, 5),
    )
    # maximise everything in columnwindow
    gui_utils.maximise_widgets(columnwindow)


def initialise_labels_and_entries(
    cfg, widget_cfg, window, key, which_case_string, *args
):
    """Add labels & entries for joint column information areas"""
    # unpack
    TEXT_FONT_NAME = widget_cfg["TEXT_FONT_NAME"]
    TEXT_FONT_SIZE = widget_cfg["TEXT_FONT_SIZE"]
    # we input a list of strings if we use this function to initialise angles (since
    # cfg["angle"] itself is a dict) => handle this particularly at textvariable
    # parameter of CTkEntry
    if type(key) is str:
        this_var = cfg[key]
    elif type(key) is list:
        this_var = cfg[key[0]][key[1]]
    joint_labels = [[] for _ in this_var]
    joint_entries = [[] for _ in this_var]
    row_counter = 0  # because we always call this in a scrollable frame
    if args:
        column_number = args
    else:
        column_number = 0
    for j, joint in enumerate(this_var):
        joint_labels[j] = ctk.CTkLabel(
            window,
            text=which_case_string + " #" + str(j + 1),
            font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
        )
        joint_labels[j].grid(row=row_counter, column=column_number, sticky="ew")
        row_counter += 1
        if type(key) is str:
            joint_entries[j] = ctk.CTkEntry(
                window, textvariable=cfg[key][j], font=(TEXT_FONT_NAME, TEXT_FONT_SIZE)
            )
        elif type(key) is list:
            joint_entries[j] = ctk.CTkEntry(
                window,
                textvariable=cfg[key[0]][key[1]][j],
                font=(TEXT_FONT_NAME, TEXT_FONT_SIZE),
            )
        # span only in width
        joint_entries[j].grid(row=row_counter, column=column_number)
        row_counter += 1
    # maximise columns
    for c in range(window.grid_size()[0]):
        window.grid_columnconfigure(c, weight=1)



import autogaita


def group_universal3Drun():
    """
    Batchrun script to run AutoGaitA Group for Results obtained with AutoGaitA Universal 3D.
    folderinfo & cfg dictionaries must be configured as explained in our documentation. See the "AutoGaitA without the GUI" section of our documentation for references to in-depth explanations to all dictionary keys (note that each key of dicts corresponds to some object in the AutoGaitA Group GUI)
    """
    # loop over legs - currently no option to do both legs in a single run
    cfg = {}
    for cfg["which_leg"] in ["left", "right"]:
        # folderinfo
        # => Note that length of folderinfo's group_names & group_dirs lists determines #    how many groups are compared.
        # => Also note that indices must correspond (i.e., idx #    1's name will be #    used for dataset stored in group_dir's idx 1)
        folderinfo = {}
        folderinfo["group_names"] = ["Young", "Old"]
        folderinfo["group_dirs"] = [
            "/Users/mahan/sciebo/Research/AutoGaitA/Human/Testing2/Young/",
            "/Users/mahan/sciebo/Research/AutoGaitA/Human/Testing2/Old/",
        ]
        folderinfo["results_dir"] = (
            "/Users/mahan/sciebo/Research/AutoGaitA/Human/Testing2/Group/"
            + cfg["which_leg"]
            + " leg/"
        )
        folderinfo["load_dir"] = ""
        # cfg
        cfg["do_permtest"] = True
        cfg["do_anova"] = True
        cfg["permutation_number"] = 10000
        cfg["PCA_n_components"] = 3
        cfg["PCA_custom_scatter_PCs"] = ""
        cfg["PCA_bins"] = ""
        cfg["PCA_save_3D_video"] = False
        cfg["stats_threshold"] = 0.05
        cfg["plot_SE"] = False
        cfg["color_palette"] = "viridis"
        cfg["legend_outside"] = True
        cfg["dont_show_plots"] = True
        cfg["anova_design"] = "Mixed ANOVA"
        cfg["PCA_variables"] = [
            # "Midfoot, " + cfg["which_leg"] + " Z",
            # "Ankle, " + cfg["which_leg"] + " Z",
            # "Knee, " + cfg["which_leg"] + " Z",
            # "Hip, " + cfg["which_leg"] + " Z",
            # "Skullbase Angle",
            # "Elbow, " + cfg["which_leg"] + " Angle",
            # "Pelvis Z",
            # "Shoulder, " + cfg["which_leg"] + " Z",
        ]
        cfg["stats_variables"] = [
            "Ankle, " + cfg["which_leg"] + " Z",
            "Ankle, " + cfg["which_leg"] + " Y",
            "Ankle, " + cfg["which_leg"] + " Velocity",
            "Ankle, " + cfg["which_leg"] + " Acceleration",
            "Ankle, " + cfg["which_leg"] + " Angle",
            "Ankle, " + cfg["which_leg"] + " Angle Velocity",
            "Ankle, " + cfg["which_leg"] + " Angle Acceleration",
            # "Knee, " + cfg["which_leg"] + " Z",
            # "Shoulder, " + cfg["which_leg"] + " Angle",
            # "Skullbase Angle",
            # "Elbow, " + cfg["which_leg"] + " Angle",
        ]
        # run
        autogaita.group(folderinfo, cfg)


# %% what happens if we just hit run
if __name__ == "__main__":
    group_universal3Drun()



# ...................................  imports  ........................................
import autogaita
import pandas as pd
import numpy as np
import os
import traceback
import math
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk
import tkinter as tk
import customtkinter as ctk

# .................................  constants  ........................................
from autogaita.resources.constants import ISSUES_TXT_FILENAME, INFO_TEXT_WIDTH, TIME_COL
from autogaita.universal3D.universal3D_constants import LEGS_COLFORMAT


# ...............................  error handling  .....................................
def try_to_run_gaita(tracking_software, info, folderinfo, cfg, multirun_flag):
    """Try to run AutoGaitA for a single dataset - print and log error if there was any
    error that prevented completion of the main code

    Note
    ----
    Needs to know "which gaita" (DLC, SLEAP or Universal 3D) should be run and if "this run" is part
    of a call to one of our multiruns!
    """

    # first print some info
    line_row = "-" * INFO_TEXT_WIDTH
    empty_row = " " * INFO_TEXT_WIDTH
    software_string = ""
    if tracking_software == "DLC":
        software_string = "D E E P L A B C U T"
    elif tracking_software == "SLEAP":
        software_string = "S L E A P"
    elif tracking_software == "Universal 3D":
        software_string = "U N I V E R S A L  3 D"
    message_1 = f"A U T O G A I T A | {software_string}"
    message_2 = "Analysing"
    message_3 = info["name"]
    side_space_1 = " " * ((INFO_TEXT_WIDTH - len(message_1)) // 2)
    side_space_2 = " " * ((INFO_TEXT_WIDTH - len(message_2)) // 2)
    if len(message_3) >= INFO_TEXT_WIDTH:
        side_space_3 = ""
    else:
        side_space_3 = " " * ((INFO_TEXT_WIDTH - len(message_3)) // 2)
    # wow
    message = f"\n\n\n{line_row}\n{side_space_1}{message_1}{side_space_1}\n{empty_row}\n{side_space_2}{message_2}{side_space_2}\n{side_space_3}{message_3}{side_space_3}\n{line_row}\n"
    print(message)

    # try to run gaita
    try:
        if tracking_software == "DLC":
            autogaita.dlc(info, folderinfo, cfg)
        elif tracking_software == "SLEAP":
            autogaita.sleap(info, folderinfo, cfg)
        elif tracking_software == "Universal 3D":
            autogaita.universal3D(info, folderinfo, cfg)
        else:
            print("tracking_software has to be DLC, SLEAP or Universal 3D - try again.")
    # catch these errors (don't catch all possbile errors - bad practice!)
    except (
        KeyError,
        IndexError,
        TypeError,
        ValueError,
        FileNotFoundError,
        IOError,
        OSError,
        PermissionError,
        MemoryError,
        OverflowError,
        FloatingPointError,
        pd.errors.ParserError,
        pd.errors.DtypeWarning,
        np.linalg.LinAlgError,
    ):
        error_traceback = traceback.format_exc()  # capture traceback of error
        # modify and print message
        skip_message = (
            "* ! ! ! ! ! ! ATTENTION  PLEASE ! ! ! ! ! ! *"
            + "\n* Whoopsie - Something unexpected is wrong! *"
            + "\n* See below & check Issues.txt for details. *"
        )
        if multirun_flag:
            skip_message += (
                "\n*   We'll continue with the next dataset!   *"
                + "\n\n*********************************************"
                + "\n*               Error Details               *"
                + "\n*********************************************"
                + f"\n{error_traceback}"
            )
        else:
            skip_message += (
                "\n\n*********************************************"
                + "\n*               Error Details               *"
                + "\n*********************************************\n"
                + f"\n{error_traceback}"
            )
        print(skip_message)
        # store message
        if not os.path.exists(info["results_dir"]):
            os.makedirs(info["results_dir"])
        textfile = os.path.join(info["results_dir"], ISSUES_TXT_FILENAME)
        with open(textfile, "a") as f:
            f.write(skip_message)
    return


# ...........................  generic helper functions  ...............................
def write_issues_to_textfile(message, info):
    """If there are any issues with this data, inform the user in this file"""
    textfile = os.path.join(info["results_dir"], ISSUES_TXT_FILENAME)
    with open(textfile, "a") as f:
        f.write(message)


def print_finish(info):
    """Print that we finished this program"""
    line_row = "-" * INFO_TEXT_WIDTH
    empty_row = " " * INFO_TEXT_WIDTH
    finished_message_1 = "A U T O G A I T A | D O N E"
    finished_message_2 = "Results are here"
    side_space_1 = " " * ((INFO_TEXT_WIDTH - len(finished_message_1)) // 2)
    side_space_2 = " " * ((INFO_TEXT_WIDTH - len(finished_message_2)) // 2)
    message = (
        "\n\n\n"
        + line_row
        + "\n"
        + side_space_1
        + finished_message_1
        + side_space_2
        + "\n"
        + empty_row
        + "\n"
        + side_space_2
        + finished_message_2
        + side_space_2
        + "\n"
        + str(info["results_dir"])
        + "\n"
        + line_row
    )
    print(message)


def compute_angle(joint_angle, joint2, joint3):
    """Compute a given angle at a joint & a given timepoint"""
    # prep our boolean to check if angle was broken
    broken = False
    # 1) vector between our angle-joint and the other two joints
    v1 = (joint_angle[0] - joint2[0], joint_angle[1] - joint2[1])
    v2 = (joint_angle[0] - joint3[0], joint_angle[1] - joint3[1])
    # 2) dot product and magnitudes of both vectors
    # => dot product is sum of each index multiplied by the other vector's index
    # => magnitude is square root of all items squared and them summed together
    dot_product = v1[0] * v2[0] + v1[1] * v2[1]
    mag_v1 = math.sqrt(v1[0] ** 2 + v1[1] ** 2)
    mag_v2 = math.sqrt(v2[0] ** 2 + v2[1] ** 2)
    # 3) cosine of the angle and angle
    # => if needed: clamp to valid range to handle floating point errors
    cos_angle = dot_product / (mag_v1 * mag_v2)
    try:
        acos_angle = math.acos(cos_angle)
    except ValueError:  # cos_angle wasn't in the -1 to +1 range
        cos_angle = max(-1, min(1, cos_angle))  # clamp to valid range
        acos_angle = math.acos(cos_angle)
        broken = True
    angle = math.degrees(acos_angle)
    return angle, broken


def write_angle_warning(step, a, angles, broken_angle_idxs, info, **kwargs):
    """Write a warning to the Issues.txt file if angles were broken"""
    # Make sure you write the first WARNING message just once
    issues_file = os.path.join(info["results_dir"], ISSUES_TXT_FILENAME)
    # the file always exists in /autogaita, but not with all /tests - handling that here
    if os.path.exists(issues_file):
        with open(issues_file, "r") as f:
            issues = f.read()
    else:
        issues = ""
    if "arc cosine to -1" not in issues:
        message = (
            "\n\n***********\n! WARNING !\n***********\n"
            + "\nWe had to limit the input of the arc cosine to -1 to +1 when "
            + "computing some angles.\nThis likely is just a rounding issue with Python"
            + " but could hint at problems with your tracking, check to be sure.\n\n"
        )
    # write the which-config broke angle - message always
    message += "\n\nArc-cosine of angle fixed for this configuration:"
    if "legname" in kwargs:  # handle Universal 3D legname
        message += f"\nLeg: {kwargs['legname']}"
    time_col_idx = step.columns.get_loc(TIME_COL)
    start_time = step.iloc[broken_angle_idxs[0], time_col_idx]
    end_time = step.iloc[broken_angle_idxs[-1], time_col_idx]
    message += (
        f"\n\nAngle: {angles['name'][a]}"
        + f"\nLower Joint: {angles['lower_joint'][a]}"
        + f"\nUpper Joint: {angles['upper_joint'][a]}"
        + f"\nCycle-time: {start_time}-{end_time}s"
    )
    print(message)
    write_issues_to_textfile(message, info)


def bin_num_to_percentages(bin_num):
    """Convert bin_num to a list of percentages"""
    # smaller than 100 means we know its integers and there are no duplicates
    if bin_num < 100:
        return [int(((s + 1) / bin_num) * 100) for s in range(bin_num)]
    # special case for 100
    elif bin_num == 100:  #
        return [s for s in range(1, 101)]
    # use floats to avoid integer duplicates for more than 100 bins
    else:
        return [round((((s + 1) / bin_num) * 100), 2) for s in range(bin_num)]


def standardise_primary_joint_coordinates(data, tracking_software, info, cfg):
    """Standardises all primary joint coordinates by a fixed decimal value which is divided from all dimensions' coordinates
    => Note in Universal 3D it's just called joint
    """
    # unpack
    name = info["name"]
    coordinate_standardisation_xls = cfg["coordinate_standardisation_xls"]
    if tracking_software != "Universal 3D":
        ID_string = str(info["mouse_num"])
        run_string = str(info["run_num"])
        joints = cfg["hind_joints"]
        angles = cfg["angles"]
    else:
        ID_string = name
        joints = cfg["joints"]
        angles = cfg["angles"]
    # first test: check if all joints used in angles are present in joints
    angle_issue_message = ""
    clean_angles = {"name": [], "lower_joint": [], "upper_joint": []}
    for i in range(len(angles["name"])):
        if all(
            angle_joint in joints
            for angle_joint in [
                angles["name"][i],
                angles["lower_joint"][i],
                angles["upper_joint"][i],
            ]
        ):
            clean_angles["name"].append(angles["name"][i])
            clean_angles["lower_joint"].append(angles["lower_joint"][i])
            clean_angles["upper_joint"].append(angles["upper_joint"][i])
        else:
            if not angle_issue_message:
                angle_issue_message += (
                    "\n***********\n! WARNING !\n***********\n"
                    + "\nYou are standardising your coordinates to fixed values / ID!"
                    + "\nYou must ensure that all joints used in angle-computations "
                    + "are present in your (hind) joints list and are thus "
                    + "standardised!\nThis would otherwise lead to broken angles."
                    + "\nProblematic angles are:"
                    + "\n\n-----------------------"
                )
            angle_issue_message += (
                f"\nName: {angles['name'][i]}"
                + f"\nLower Joint: {angles['lower_joint'][i]}"
                + f"\nUpper Joint: {angles['upper_joint'][i]}"
                + "\n-----------------------"
            )
    # rare case of overwriting cfg within autogaita!
    if angle_issue_message:
        angle_issue_message += (
            "\nWe will update your cfg (!) & continue with the remaining angles!"
        )
        print(angle_issue_message)
        write_issues_to_textfile(angle_issue_message, info)
    cfg["angles"] = clean_angles
    # test if xls exists - if not quit autogaita
    # => note that some_prep will be stopped if this should return None
    if not os.path.exists(coordinate_standardisation_xls):
        message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "No coordinate standardisation xls file found at:"
            + f"\n{coordinate_standardisation_xls}!"
            + "\nFix the path or remove the value from the config (GUI) if you "
            + "do not wish you standardise coordinates."
            + "\nNote that horizontal (i.e. x- in 2D) or height (i.e. y- in 2D) "
            + "standardisation is something else!"
            + "\nCancelling AutoGaitA!"
        )
        print(message)
        write_issues_to_textfile(message, info)
        raise FileNotFoundError
    # load the file (string because of comparison in conditon)
    coord_stand_df = pd.read_excel(coordinate_standardisation_xls).astype(str)
    if not all(coord_stand_df.columns.isin(["ID", "Run", "Standardisation Value"])):
        message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + "The coordinate standardisation Excel file does not have the correct "
            + "column names - refer to our Template file!"
            + "\nPlease check the xls file and try again."
            + "\nCancelling AutoGaitA!"
        )
        print(message)
        write_issues_to_textfile(message, info)
        raise ValueError
    # extract the row we need
    if tracking_software != "Universal 3D":
        condition = (coord_stand_df["ID"] == ID_string) & (
            coord_stand_df["Run"] == run_string
        )
    else:
        condition = coord_stand_df["ID"] == ID_string
    coord_stand_df = coord_stand_df[condition]
    if len(coord_stand_df) != 1:
        message = "\n******************\n! CRITICAL ERROR !\n******************\n"
        if len(coord_stand_df) == 0:
            message += (
                f"Unable to find {name} in the coordinate standardisation xls file!"
            )
        elif len(coord_stand_df) > 1:
            message += f"Found multiple entries for {name} in the coordinate standardisation xls file!"
        message += (
            "\nPlease check the xls file and try again." + "\nCancelling AutoGaitA!"
        )
        print(message)
        write_issues_to_textfile(message, info)
        raise ValueError
    # extract standardisation value from xls
    try:
        # if-len-lines in error-message block above ensure that the line below is
        # indexing correctly using iloc[0, 2] - condition is always of len==1!
        coordinate_standardisation_value = float(coord_stand_df.iloc[0, 2])
    except ValueError:
        message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + f"Unable to convert standardisation value for {name} to a float!"
            + "\nPlease check the xls file and try again."
            + "\nCancelling AutoGaitA!"
        )
        print(message)
        write_issues_to_textfile(message, info)
        raise ValueError
    # ensure that value is equal to or larger than 1 (otherwise we would increase vals)
    if coordinate_standardisation_value < 1:
        message = (
            "\n******************\n! CRITICAL ERROR !\n******************\n"
            + f"Standardisation value for {name} is smaller than 1!"
            + "\nPlease read the doc for why this is not allowed."
            + "\nIf you do not want to standardise this dataset, use a value of 1."
            + "\nCancelling AutoGaitA!"
        )
        print(message)
        write_issues_to_textfile(message, info)
        raise ValueError
    # all tests are passed - standardise coordinates
    # => if we are in 3D we have to check for bodyside-specificity, add all cols
    #    (joint + coord) to a list and use the list for looping when standardising
    if tracking_software != "Universal 3D":
        coordinates = ("x", "y")
    else:
        coordinates = ("X", "Y", "Z")
    cols_to_standardise = []
    for joint in joints:
        for coord in coordinates:
            central_joint = joint + coord
            side_specific_joint = joint + LEGS_COLFORMAT[0] + coord
            if central_joint in data.columns:
                cols_to_standardise.append(central_joint)
            if side_specific_joint in data.columns:
                for leg in LEGS_COLFORMAT:
                    cols_to_standardise.append(joint + leg + coord)
    data.loc[:, cols_to_standardise] = (
        data.loc[:, cols_to_standardise] / coordinate_standardisation_value
    )
    return data, cfg


# ................................  plot panel  ........................................
class PlotPanel:
    def __init__(self, fg_color, hover_color):
        self.figures = []
        self.current_fig_index = 0
        self.fg_color = fg_color
        self.hover_color = hover_color

    # .........................  loading screen  ................................
    def build_plot_panel_loading_screen(self):
        """Builds a loading screen that is shown while plots are generated"""
        # Build window
        self.loading_screen = ctk.CTkToplevel()
        self.loading_screen.title("Loading...")
        self.loading_screen.geometry("300x300")
        self.loading_label_strings = [
            "Plots are generated, please wait.",
            "Plots are generated, please wait..",
            "Plots are generated, please wait...",
        ]
        self.loading_label = ctk.CTkLabel(
            self.loading_screen, text=self.loading_label_strings[0]
        )
        self.loading_label.pack(pady=130, padx=40, anchor="w")

        # Animate the text
        self.animate(counter=1)

    # Cycle through loading labels to animate the loading screen
    def animate(self, counter):
        self.loading_label.configure(text=self.loading_label_strings[counter])
        self.loading_screen.after(
            500, self.animate, (counter + 1) % len(self.loading_label_strings)
        )

    def destroy_plot_panel_loading_screen(self):
        self.loading_screen.destroy()

    # .........................  plot panel   ................................
    def build_plot_panel(self):
        """Creates the window/"panel" in which the plots are shown"""
        # Set up of the plotpanel
        ctk.set_appearance_mode("dark")  # Modes: system (default), light, dark
        ctk.set_default_color_theme("green")  # Themes: blue , dark-blue, green
        self.plotwindow = ctk.CTkToplevel()
        self.plotwindow.title(
            f"AutoGaitA Figure {self.current_fig_index + 1}/{len(self.figures)}"
        )

        # Set size to 50% of screen
        screen_width = self.plotwindow.winfo_screenwidth()
        window_width = int(screen_width * 0.5)
        # 0.75 to gain a ration of 1.333 (that of matplotlib figures) and 1.05 for toolbar + buttons
        window_height = window_width * 0.75 * 1.05
        self.plotwindow.geometry(f"{window_width}x{window_height}")

        # Adjust figures for the plot panel
        for fig in self.figures:
            # dpi adjusted to increase visibilty/readability
            fig.set_dpi(100)
            # constrained layout to adjust margins within the figure
            # => note: in case there are a lot of steps in one run (-> the legend is
            #          super long) the figure won't be displayed properly.
            fig.set_constrained_layout(True)

        # Initialize the plot panel with the first figure
        self.plot_panel = FigureCanvasTkAgg(
            self.figures[self.current_fig_index], master=self.plotwindow
        )  # index used for buttons
        self.plot_panel.get_tk_widget().grid(
            row=0, column=0, padx=10, pady=10, sticky="nsew"
        )

        # Create toolbar frame and place it in the middle row
        self.toolbar_frame = tk.Frame(self.plotwindow)
        self.toolbar_frame.grid(row=1, column=0, sticky="ew")

        self.toolbar = NavigationToolbar2Tk(self.plot_panel, self.toolbar_frame)
        self.toolbar.update()

        # Create navigation buttons frame
        self.button_frame = tk.Frame(self.plotwindow)
        self.button_frame.grid(row=2, column=0, sticky="ew")

        self.prev_button = ctk.CTkButton(
            self.button_frame,
            text="<< Previous",
            fg_color=self.fg_color,
            hover_color=self.hover_color,
            command=self.show_previous,
        )
        self.next_button = ctk.CTkButton(
            self.button_frame,
            text="Next >>",
            fg_color=self.fg_color,
            hover_color=self.hover_color,
            command=self.show_next,
        )
        self.prev_button.grid(row=0, column=0, sticky="ew")
        self.next_button.grid(row=0, column=1, sticky="ew")

        self.button_frame.grid_columnconfigure(0, weight=1)
        self.button_frame.grid_columnconfigure(1, weight=1)

        # Configure grid layout
        self.plotwindow.grid_rowconfigure(0, weight=1)
        self.plotwindow.grid_rowconfigure(1, weight=0)
        self.plotwindow.grid_rowconfigure(2, weight=0)
        self.plotwindow.grid_columnconfigure(0, weight=1)

    def show_previous(self):
        if self.current_fig_index > 0:
            self.current_fig_index -= 1
            self.update_plot_and_toolbar()

    def show_next(self):
        if self.current_fig_index < len(self.figures) - 1:
            self.current_fig_index += 1
            self.update_plot_and_toolbar()

    def update_plot_and_toolbar(self):
        # Clear the current plot panel
        self.plot_panel.get_tk_widget().grid_forget()

        # Update the plot panel with the new figure
        self.plot_panel = FigureCanvasTkAgg(
            self.figures[self.current_fig_index], master=self.plotwindow
        )
        self.plot_panel.get_tk_widget().grid(row=0, column=0, sticky="nsew")
        self.plot_panel.draw()

        # Destroy toolbar and create a new one
        # (This has to be done, otherwise the toolbar won't function for a new plot)
        self.toolbar.destroy()
        self.toolbar = NavigationToolbar2Tk(self.plot_panel, self.toolbar_frame)
        self.toolbar.update()

        # Update title
        self.plotwindow.title(
            f"AutoGaitA Plot Panel {self.current_fig_index + 1}/{len(self.figures)}"
        )

    def destroy_plot_panel(self):
        # Needed if no SCs after checks
        self.loading_screen.destroy()



# %%...............  SHARED 2D WINDOW - ADVANCED CONFIGURATION  ........................
import autogaita.gui.gui_utils as gui_utils
import autogaita.gui.gaita_widgets as gaita_widgets
import customtkinter as ctk


def build_cfg_window(root, cfg, widget_cfg, root_dimensions):
    """Build advanced configuration window"""
    # unpack root window dimensions
    w = root_dimensions[0]
    h = root_dimensions[1]
    # build window
    cfg_window = ctk.CTkToplevel(root)
    cfg_window.title("Advanced Configuration")
    cfg_window.geometry(f"{int(w / 2)}x{h}+{int(w / 4)}+0")
    gui_utils.fix_window_after_its_creation(cfg_window)

    # unpack
    FG_COLOR = widget_cfg["FG_COLOR"]
    HOVER_COLOR = widget_cfg["HOVER_COLOR"]
    HEADER_FONT_NAME = widget_cfg["HEADER_FONT_NAME"]
    HEADER_FONT_SIZE = widget_cfg["HEADER_FONT_SIZE"]
    CLOSE_COLOR = widget_cfg["CLOSE_COLOR"]
    CLOSE_HOVER_COLOR = widget_cfg["CLOSE_HOVER_COLOR"]
    TEXT_FONT_NAME = widget_cfg["TEXT_FONT_NAME"]
    ADV_CFG_TEXT_FONT_SIZE = widget_cfg["ADV_CFG_TEXT_FONT_SIZE"]
    COLOR_PALETTES_LIST = widget_cfg["COLOR_PALETTES_LIST"]

    # advanced analysis header
    adv_cfg_analysis_header_label = gaita_widgets.header_label(
        cfg_window,
        "Analysis",
        widget_cfg,
    )
    adv_cfg_analysis_header_label.grid(
        row=0, column=0, rowspan=2, columnspan=2, sticky="nsew"
    )

    # x & y threshold for rejecting SCs
    threshold_string = "What criteria (in pixels) to use for rejecting step cycles?"
    thresh_label = ctk.CTkLabel(
        cfg_window, text=threshold_string, font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE)
    )
    thresh_label.grid(row=2, column=0, columnspan=2, sticky="ew")
    x_threshold_string = "Along x-dimension:"
    x_thresh_label = ctk.CTkLabel(
        cfg_window,
        text=x_threshold_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    x_thresh_label.grid(row=3, column=0, sticky="e")
    x_thresh_entry = ctk.CTkEntry(
        cfg_window,
        textvariable=cfg["x_sc_broken_threshold"],
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    x_thresh_entry.grid(row=3, column=1, sticky="w")
    # y threshold for rejecting SCs
    y_thresh_string = "Along y-dimension:"
    y_thresh_label = ctk.CTkLabel(
        cfg_window, text=y_thresh_string, font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE)
    )
    y_thresh_label.grid(row=4, column=0, sticky="e")
    y_thresh_entry = ctk.CTkEntry(
        cfg_window,
        textvariable=cfg["y_sc_broken_threshold"],
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    y_thresh_entry.grid(row=4, column=1, sticky="w")

    # acceleration label
    acceleration_string = "Analyse (i.e. plot & export) accelerations for:"
    acceleration_label = ctk.CTkLabel(
        cfg_window,
        text=acceleration_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    acceleration_label.grid(row=5, column=0, columnspan=2)
    # x acceleration
    x_accel_box = ctk.CTkCheckBox(
        cfg_window,
        text="x-coordinates",
        variable=cfg["x_acceleration"],
        onvalue=True,
        offvalue=False,
        hover_color=HOVER_COLOR,
        fg_color=FG_COLOR,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    x_accel_box.grid(row=6, column=0, sticky="e")
    # angular acceleration
    angular_accel_box = ctk.CTkCheckBox(
        cfg_window,
        text="angles",
        variable=cfg["angular_acceleration"],
        onvalue=True,
        offvalue=False,
        hover_color=HOVER_COLOR,
        fg_color=FG_COLOR,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    angular_accel_box.grid(row=6, column=1, sticky="w")

    # y standardisation at each step cycle separately
    standardise_y_at_SC_level_box = gaita_widgets.checkbox(
        cfg_window,
        "Standardise y-coordinates separately for all step cycles",
        cfg["standardise_y_at_SC_level"],
        widget_cfg,
        adv_cfg_textsize=True,
    )
    standardise_y_at_SC_level_box.grid(row=7, column=0, columnspan=2)

    # y standardisation to a specific joint not global minimum
    standardise_y_to_joint_box = gaita_widgets.checkbox(
        cfg_window,
        "Standardise y to a joint instead of to global minimum",
        cfg["standardise_y_to_a_joint"],
        widget_cfg,
        adv_cfg_textsize=True,
    )
    standardise_y_to_joint_box.configure(
        command=lambda: gui_utils.change_widget_state_based_on_checkbox(
            cfg, "standardise_y_to_a_joint", y_standardisation_joint_entry
        ),
    )
    standardise_y_to_joint_box.grid(row=8, column=0, columnspan=2)

    # y standardisation joint string label & entry
    y_standardisation_joint_label, y_standardisation_joint_entry = (
        gaita_widgets.label_and_entry_pair(
            cfg_window,
            "Y-standardisation joint:",
            cfg["y_standardisation_joint"][0],
            widget_cfg,
            adv_cfg_textsize=True,
        )
    )
    y_standardisation_joint_label.grid(row=9, column=0, sticky="e")
    y_standardisation_joint_entry.grid(row=9, column=1, sticky="w")
    # to initialise the widget correctly, run this function once
    gui_utils.change_widget_state_based_on_checkbox(
        cfg, "standardise_y_to_a_joint", y_standardisation_joint_entry
    )

    # analyse average x coordinates
    analyse_average_x_box = gaita_widgets.checkbox(
        cfg_window,
        "Analyse x-coordinate averages",
        cfg["analyse_average_x"],
        widget_cfg,
        adv_cfg_textsize=True,
    )
    analyse_average_x_box.configure(
        command=lambda: gui_utils.change_widget_state_based_on_checkbox(
            cfg, "analyse_average_x", standardise_x_coordinates_box
        ),
    )
    analyse_average_x_box.grid(row=10, column=0)

    # standardise x coordinates
    standardise_x_coordinates_box = gaita_widgets.checkbox(
        cfg_window,
        "Standardise x-coordinates",
        cfg["standardise_x_coordinates"],
        widget_cfg,
        adv_cfg_textsize=True,
    )
    standardise_x_coordinates_box.configure(
        command=lambda: gui_utils.change_widget_state_based_on_checkbox(
            cfg, "standardise_x_coordinates", x_standardisation_joint_entry
        ),
    )
    standardise_x_coordinates_box.grid(row=10, column=1)
    gui_utils.change_widget_state_based_on_checkbox(
        cfg, "analyse_average_x", standardise_x_coordinates_box
    )

    # x standardisation joint string label & entry
    x_standardisation_joint_label, x_standardisation_joint_entry = (
        gaita_widgets.label_and_entry_pair(
            cfg_window,
            "X-standardisation joint:",
            cfg["x_standardisation_joint"][0],
            widget_cfg,
            adv_cfg_textsize=True,
        )
    )
    x_standardisation_joint_label.grid(row=11, column=0, sticky="e")
    x_standardisation_joint_entry.grid(row=11, column=1, sticky="w")
    gui_utils.change_widget_state_based_on_checkbox(
        cfg, "standardise_x_coordinates", x_standardisation_joint_entry
    )

    # invert y-axis
    if "invert_y_axis" in cfg.keys():
        invert_y_axis_box = gaita_widgets.checkbox(
            cfg_window,
            "Invert y-axis",
            cfg["invert_y_axis"],
            widget_cfg,
            adv_cfg_textsize=True,
        )
        invert_y_axis_box.grid(row=12, column=0, columnspan=2)

    # Standardise all (primary) joint coordinates by a fixed decimal value
    coordinate_standardisation_xls_label, coordinate_standardisation_xls_entry = (
        gaita_widgets.label_and_entry_pair(
            cfg_window,
            "Excel file for primary-joint coordinate standardisation:",
            cfg["coordinate_standardisation_xls"],
            widget_cfg,
            adv_cfg_textsize=True,
        )
    )
    coordinate_standardisation_xls_label.grid(row=13, column=0, columnspan=2)
    coordinate_standardisation_xls_entry.grid(row=14, column=0, columnspan=2)

    #  .............................  advanced output  .................................
    # advanced analysis header
    adv_cfg_output_header_label = gaita_widgets.header_label(
        cfg_window,
        "Output",
        widget_cfg,
    )
    adv_cfg_output_header_label.grid(
        row=15, column=0, rowspan=2, columnspan=2, sticky="nsew"
    )

    # number of hindlimb (primary) joints to plot
    plot_joint_num__label, plot_joint_num_entry = gaita_widgets.label_and_entry_pair(
        cfg_window,
        "Number of primary joints to plot in detail:",
        cfg["plot_joint_number"],
        widget_cfg,
        adv_cfg_textsize=True,
    )
    plot_joint_num__label.grid(row=17, column=0, columnspan=2)
    plot_joint_num_entry.grid(row=18, column=0, columnspan=2)

    # save to xls
    save_to_xls_box = gaita_widgets.checkbox(
        cfg_window,
        "Save results as .xlsx instead of .csv files",
        cfg["save_to_xls"],
        widget_cfg,
        adv_cfg_textsize=True,
    )
    save_to_xls_box.grid(row=19, column=0, columnspan=2)

    # plot SE
    plot_SE_box = gaita_widgets.checkbox(
        cfg_window,
        "Use standard error instead of standard deviation for plots",
        cfg["plot_SE"],
        widget_cfg,
        adv_cfg_textsize=True,
    )
    plot_SE_box.grid(row=20, column=0, columnspan=2)

    # color palette
    color_palette_string = "Choose figures' color palette"
    color_palette_label = ctk.CTkLabel(
        cfg_window,
        text=color_palette_string,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    color_palette_label.grid(row=21, column=0, columnspan=2)
    color_palette_entry = ctk.CTkOptionMenu(
        cfg_window,
        values=COLOR_PALETTES_LIST,
        variable=cfg["color_palette"],
        fg_color=FG_COLOR,
        button_color=FG_COLOR,
        button_hover_color=HOVER_COLOR,
        font=(TEXT_FONT_NAME, ADV_CFG_TEXT_FONT_SIZE),
    )
    color_palette_entry.grid(row=22, column=0, columnspan=2)

    # legend outside
    legend_outside_checkbox = gaita_widgets.checkbox(
        cfg_window,
        "Plot legends outside of figures' panels",
        cfg["legend_outside"],
        widget_cfg,
        adv_cfg_textsize=True,
    )
    legend_outside_checkbox.grid(row=23, column=0, columnspan=2)

    # results dir
    results_dir_label, results_dir_entry = gaita_widgets.label_and_entry_pair(
        cfg_window,
        "Save Results subfolders to directory below instead of to data's:",
        cfg["results_dir"],
        widget_cfg,
        adv_cfg_textsize=True,
    )
    results_dir_label.grid(row=24, column=0, columnspan=2)
    results_dir_entry.grid(row=25, column=0, columnspan=2)

    # done button
    adv_cfg_done_button = ctk.CTkButton(
        cfg_window,
        text="I am done, update cfg.",
        fg_color=CLOSE_COLOR,
        hover_color=CLOSE_HOVER_COLOR,
        font=(HEADER_FONT_NAME, HEADER_FONT_SIZE),
        command=lambda: cfg_window.destroy(),
    )
    adv_cfg_done_button.grid(
        row=26, column=0, columnspan=2, rowspan=2, sticky="nsew", padx=10, pady=(10, 5)
    )
    # maximise widgets
    cfg_window.columnconfigure(list(range(2)), weight=1, uniform="Silent_Creme")
    cfg_window.rowconfigure(list(range(25)), weight=1, uniform="Silent_Creme")



import autogaita


def group_dlcrun():
    """
    Batchrun script to run AutoGaitA Group for Results obtained with AutoGaitA DLC.
    folderinfo & cfg dictionaries must be configured as explained in our documentation. See the "AutoGaitA without the GUI" section of our documentation for references to in-depth explanations to all dictionary keys (note that each key of dicts corresponds to some object in the AutoGaitA Group GUI)
    """
    # folderinfo
    # => Note that length of folderinfo's group_names & group_dirs lists determines how #    many groups are compared.
    # => Also note that indices must correspond (i.e., idx #    1's name will be used #    for dataset stored in group_dir's idx 1)
    folderinfo = {}
    folderinfo["group_names"] = [
        # RM Dataset
        "5mm",
        "12mm",
        "25mm",
        # Interaction Issue Dataset
        # "Control",
        # "Silenced",
        # Testing Dataset
        # "one",
        # "two",
    ]
    folderinfo["group_dirs"] = [
        # --------------------------
        # Full Dataset (39 subjects)
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Control Full",
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Silenced Full/",
        # --------------------------
        # Reduced Dataset (5 subjects)
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Control",
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Silenced/",
        # --------------------------
        # Testing Dataset
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Testing/Group 1/",
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Testing/Group 2/",
        # --------------------------
        # RM Dataset (example data in repo)
        "/Users/mahan/sciebo/PythonCode/gaita_repo/example data/5mm/Results/",
        "/Users/mahan/sciebo/PythonCode/gaita_repo/example data/12mm/Results/",
        "/Users/mahan/sciebo/PythonCode/gaita_repo/example data/25mm/Results/",
    ]
    folderinfo["results_dir"] = (
        "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/example_data_results/"
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Sanity Test Interaction Graziana/Result 10 joints/Mahan Results/"
        # "/Users/mahan/sciebo/Research/AutoGaitA/Mouse/Testing/GroupResults/"
    )
    folderinfo["load_dir"] = ""
    # cfg
    cfg = {}
    cfg["do_permtest"] = False  # True
    cfg["do_anova"] = True
    cfg["permutation_number"] = 100
    cfg["PCA_n_components"] = 6
    # cfg["PCA_n_components"] = 0.33
    # cfg["PCA_custom_scatter_PCs"] = "4,5,6;4,5;2,4,6"
    cfg["PCA_custom_scatter_PCs"] = ""
    cfg["PCA_save_3D_video"] = False  # True
    cfg["PCA_bins"] = ""  # "0-10,24,50-75"
    cfg["stats_threshold"] = 0.05
    cfg["plot_SE"] = False
    cfg["color_palette"] = "viridis"
    cfg["dont_show_plots"] = True
    cfg["legend_outside"] = True
    cfg["which_leg"] = "left"
    cfg["anova_design"] = "RM ANOVA"
    # cfg["anova_design"] = "Mixed ANOVA"
    cfg["permutation_number"] = 100
    cfg["PCA_variables"] = [
        # "Hind paw tao y",
        # "Ankle y",
        # "Knee y",
        # "Ankle Angle",
        # "Knee Angle",
        # "Nose x",
        # "Knee y",
        # "Knee x",
        # "Knee Velocity",
        # "Knee Acceleration",
        # "Knee Angle",
        # "Knee Angle Velocity",
        # "Knee Angle Acceleration",
        # "Hip Angle",
    ]
    cfg["stats_variables"] = [
        "Hind paw tao y",
        # "Ankle y",
        "Knee y",
        # "Ankle Angle",
        "Knee Angle",
        # "Nose x",
        # "Knee y",
        "Knee x",
        # "Knee Velocity",
        # "Knee Acceleration",
        # "Knee Angle",
        # "Knee Angle Velocity",
        # "Knee Angle Acceleration",
        # "Hip Angle",
        # "Elbow Angle"
    ]
    # run
    autogaita.group(folderinfo, cfg)


# %% what happens if we just hit run
if __name__ == "__main__":
    group_dlcrun()
